//
//  CameraKitFlutterView.swift
//  camerakit
//
//  Created by faranegar on 9/6/20.
//

import Foundation
import AVFoundation
import MLKitBarcodeScanning
import MLKitCommon
import MLKitVision


@available(iOS 10.0, *)
class CameraKitFlutterView : NSObject, FlutterPlatformView, AVCaptureVideoDataOutputSampleBufferDelegate, AVCapturePhotoCaptureDelegate{
    let channel: FlutterMethodChannel
    let frame: CGRect

    var hasBarcodeReader:Bool!
    var flashMode:AVCaptureDevice.FlashMode!
    
    var previewView : UIView!
    var videoDataOutput: AVCaptureVideoDataOutput!
    var videoDataOutputQueue: DispatchQueue!
    
    var photoOutput: AVCapturePhotoOutput?
    var previewLayer:AVCaptureVideoPreviewLayer!
    var captureDevice : AVCaptureDevice!
    let session = AVCaptureSession()
    var barcodeScanner:BarcodeScanner!
    var flutterResultTakePicture:FlutterResult!
    
    init(registrar: FlutterPluginRegistrar, viewId: Int64, frame: CGRect) {
         self.channel = FlutterMethodChannel(name: "plugins/camera_kit_" + String(viewId), binaryMessenger: registrar.messenger())
        self.frame = frame
         

    
     }
    
    
    
    public func setMethodHandler() {
        self.channel.setMethodCallHandler({ (FlutterMethodCall,  FlutterResult) in
                let args = FlutterMethodCall.arguments
                let myArgs = args as? [String: Any]
                if FlutterMethodCall.method == "requestPermission" {
                    FlutterResult(true)
                } else if FlutterMethodCall.method == "initCamera" {
            
                    self.initCamera(hasBarcodeReader: (myArgs?["hasBarcodeReader"] as! Bool),
                                    flashMode: (myArgs?["flashMode"] ) as! String)
                } else if FlutterMethodCall.method == "resumeCamera" {
                    self.beginSession()
            }
                else if FlutterMethodCall.method == "pauseCamera" {
                    self.stopCamera()
                }
            else if FlutterMethodCall.method == "changeFlashMode" {
//                    self.initCamera()
                    self.setFlashMode(flashMode: (myArgs?["flashMode"] ) as! String)
                    self.changeFlashMode()
            }
             else if FlutterMethodCall.method == "takePicture" {
                    self.flutterResultTakePicture = FlutterResult
                    self.takePicture()
                        }
            })
    }
    
    func changeFlashMode() {
       
        if(self.hasBarcodeReader) {
            do{
               if (captureDevice.hasTorch)
                   {
                       try captureDevice.lockForConfiguration()
                    captureDevice.torchMode = (self.flashMode == .auto) ?(.auto):(self.flashMode == .on ? (.on) : (.off))
                       captureDevice.flashMode = self.flashMode
                       captureDevice.unlockForConfiguration()
                   }
                }catch{
                   //DISABEL FLASH BUTTON HERE IF ERROR
                   print("Device tourch Flash Error ");
               }
          
        }
    }
    
    func setFlashMode(flashMode: String) {
        if flashMode == "A" {
            self.flashMode = .auto
                  } else if flashMode == "O" {
            self.flashMode = .on
                  } else if flashMode == "F"{
            self.flashMode = .off
                  }
    }
    
    func view() -> UIView {
        if previewView == nil {
        self.previewView = UIView(frame: frame)
         previewView.contentMode = UIView.ContentMode.scaleAspectFit
        }
        return previewView
    }
    
    func initCamera(hasBarcodeReader: Bool, flashMode: String) {
        self.hasBarcodeReader = hasBarcodeReader
        setFlashMode(flashMode: flashMode)
        if hasBarcodeReader == true{
            let format = BarcodeFormat.all
                let barcodeOptions = BarcodeScannerOptions(formats: format)

                // Create a barcode scanner.
               barcodeScanner = BarcodeScanner.barcodeScanner(options: barcodeOptions)
        }
        DispatchQueue.main.asyncAfter(deadline: .now() + 0.5) {
            self.setupAVCapture()
        }
        
 
    }
    
    @available(iOS 10.0, *)
    func setupAVCapture(){
          session.sessionPreset = AVCaptureSession.Preset.vga640x480
          guard let device = AVCaptureDevice
          .default(AVCaptureDevice.DeviceType.builtInWideAngleCamera,
                   for: .video,
                   position: AVCaptureDevice.Position.back) else {
                              return
          }
          captureDevice = device
    
        changeFlashMode()
          beginSession()
      }
    
    private func flashAuto()
    {
        do{
            if (captureDevice.hasTorch)
            {
                try captureDevice.lockForConfiguration()
                captureDevice.torchMode = .auto
                captureDevice.flashMode = .auto
                captureDevice.unlockForConfiguration()
            }
        }catch{
            //DISABEL FLASH BUTTON HERE IF ERROR
            print("Device tourch Flash Error ");
        }
    }
    
    private func flashOn()
    {
        do{
            if (captureDevice.hasTorch)
            {
                try captureDevice.lockForConfiguration()
                captureDevice.torchMode = .on
                captureDevice.flashMode = .on
                captureDevice.unlockForConfiguration()
            }
        }catch{
            //DISABEL FLASH BUTTON HERE IF ERROR
            print("Device tourch Flash Error ");
        }
    }
    
    private func flashOff()
       {
           do{
               if (captureDevice.hasTorch){
                   try captureDevice.lockForConfiguration()
                   captureDevice.torchMode = .off
                   captureDevice.flashMode = .off
                   captureDevice.unlockForConfiguration()
               }
           }catch{
               //DISABEL FLASH BUTTON HERE IF ERROR
               print("Device tourch Flash Error ");
           }
       }
    
    func beginSession(){
        var deviceInput: AVCaptureDeviceInput!

        
        do {
            deviceInput = try AVCaptureDeviceInput(device: captureDevice)
            guard deviceInput != nil else {
                print("error: cant get deviceInput")
                return
            }
            
    
            

            if self.session.canAddInput(deviceInput){
                self.session.addInput(deviceInput)
            }

            if(hasBarcodeReader) {
                videoDataOutput = AVCaptureVideoDataOutput()
                videoDataOutput.alwaysDiscardsLateVideoFrames=true
                videoDataOutputQueue = DispatchQueue(label: "VideoDataOutputQueue")
                videoDataOutput.setSampleBufferDelegate(self, queue:self.videoDataOutputQueue)
                if session.canAddOutput(videoDataOutput!){
                             session.addOutput(videoDataOutput!)
                 }
                videoDataOutput.connection(with: .video)?.isEnabled = true

            }
            else {
                photoOutput = AVCapturePhotoOutput()
                    photoOutput?.setPreparedPhotoSettingsArray([AVCapturePhotoSettings(format: [AVVideoCodecKey : AVVideoCodecJPEG])], completionHandler: nil)
                if session.canAddOutput(photoOutput!){
                    session.addOutput(photoOutput!)
                }
            }



            previewLayer = AVCaptureVideoPreviewLayer(session: self.session)
            previewLayer.videoGravity = AVLayerVideoGravity.resizeAspect

            let rootLayer :CALayer = self.previewView.layer
            rootLayer.masksToBounds = true
            previewLayer.frame = rootLayer.bounds
            rootLayer.addSublayer(self.previewLayer)
            session.startRunning()
        } catch let error as NSError {
            deviceInput = nil
            print("error: \(error.localizedDescription)")
        }
    }
    
    func stopCamera(){
        session.stopRunning()
    }

    
     private func currentUIOrientation() -> UIDeviceOrientation {
        let deviceOrientation = { () -> UIDeviceOrientation in
          switch UIApplication.shared.statusBarOrientation {
          case .landscapeLeft:
            return .landscapeRight
          case .landscapeRight:
            return .landscapeLeft
          case .portraitUpsideDown:
            return .portraitUpsideDown
          case .portrait, .unknown:
            return .portrait
          @unknown default:
            fatalError()
          }
        }
        guard Thread.isMainThread else {
          var currentOrientation: UIDeviceOrientation = .portrait
          DispatchQueue.main.sync {
            currentOrientation = deviceOrientation()
          }
          return currentOrientation
        }
        return deviceOrientation()
      }
    
    
    public func imageOrientation(
      fromDevicePosition devicePosition: AVCaptureDevice.Position = .back
    ) -> UIImage.Orientation {
      var deviceOrientation = UIDevice.current.orientation
      if deviceOrientation == .faceDown || deviceOrientation == .faceUp
        || deviceOrientation
          == .unknown
      {
        deviceOrientation = currentUIOrientation()
      }
      switch deviceOrientation {
      case .portrait:
        return devicePosition == .front ? .leftMirrored : .right
      case .landscapeLeft:
        return devicePosition == .front ? .downMirrored : .up
      case .portraitUpsideDown:
        return devicePosition == .front ? .rightMirrored : .left
      case .landscapeRight:
        return devicePosition == .front ? .upMirrored : .down
      case .faceDown, .faceUp, .unknown:
        return .up
      @unknown default:
        fatalError()
      }
    }
    func saveImage(image: UIImage) -> Bool {
        guard let data = image.jpegData(compressionQuality: 1) ?? image.pngData() else {
            return false
        }
        guard let directory = try? FileManager.default.url(for: .documentDirectory, in: .userDomainMask, appropriateFor: nil, create: false) as NSURL else {
            return false
        }
        do {
            try data.write(to: directory.appendingPathComponent("pic.jpg")!)
            flutterResultTakePicture(directory.path!  + "/pic.jpg")
            //print(directory)
            return true
        } catch {
            print(error.localizedDescription)
                        flutterResultTakePicture(FlutterError(code: "-103", message: error.localizedDescription, details: nil))
            return false
        }
    }
    func takePicture() {
        let settings = AVCapturePhotoSettings()
        settings.flashMode = self.flashMode
        photoOutput?.capturePhoto(with: settings, delegate:self)
    }
    
    public func photoOutput(_ captureOutput: AVCapturePhotoOutput, didFinishProcessingPhoto photoSampleBuffer: CMSampleBuffer?, previewPhoto previewPhotoSampleBuffer: CMSampleBuffer?,
                        resolvedSettings: AVCaptureResolvedPhotoSettings, bracketSettings: AVCaptureBracketedStillImageSettings?, error: Swift.Error?) {
        if let error = error { //self.photoCaptureCompletionBlock?(nil, error)
            flutterResultTakePicture(FlutterError(code: "-101", message: error.localizedDescription, details: nil))
        }
            
        else if let buffer = photoSampleBuffer, let data = AVCapturePhotoOutput.jpegPhotoDataRepresentation(forJPEGSampleBuffer: buffer, previewPhotoSampleBuffer: nil),
            let image = UIImage(data: data) {
            
            self.saveImage(image: image)
        }
            
        else {
            //error
//            self.photoCaptureCompletionBlock?(nil, CameraControllerError.unknown)
                        flutterResultTakePicture(FlutterError(code: "-102", message: "Unknown error", details: nil))
        }
    }
    

    func captureOutput(_ output: AVCaptureOutput, didOutput sampleBuffer: CMSampleBuffer, from connection: AVCaptureConnection) {
        

        
        // do stuff here
        if barcodeScanner != nil {
            let visionImage = VisionImage(buffer: sampleBuffer)
            let orientation = imageOrientation(
              fromDevicePosition: .back
            )
            visionImage.orientation = orientation
            var barcodes: [Barcode]
            do {
                barcodes = try self.barcodeScanner.results(in: visionImage)
            } catch let error {
              print("Failed to scan barcodes with error: \(error.localizedDescription).")
              return
            }
            
            guard !barcodes.isEmpty else {
               //print("Barcode scanner returrned no results.")
               return
             }
            
            for barcode in barcodes {
                barcodeRead(barcode: barcode.rawValue!)
            }
        }
        
    }
    
    func barcodeRead(barcode: String) {
        channel.invokeMethod("onBarcodeRead", arguments: barcode)
    }
    
}
